/*
 * xsltproc.c: user program for the XSL Transformation 1.0 engine
 *
 * See Copyright for the status of this software.
 *
 * daniel@veillard.com
 */

//#include <libxslt/libxslt.h>
//#include <libexslt/exslt.h>
#include <string.h>
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdlib.h>
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#include <libxml/xmlmemory.h>
#include <libxml/debugXML.h>
#include <libxml/HTMLtree.h>
#include <libxml/xmlIO.h>
#ifdef LIBXML_XINCLUDE_ENABLED
#include <libxml/xinclude.h>
#endif
#ifdef LIBXML_CATALOG_ENABLED
#include <libxml/catalog.h>
#endif
#include <libxml/parserInternals.h>

#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>
#include <libxslt/extensions.h>

#include <libexslt/exslt.h>


#ifndef HAVE_STAT
#  ifdef HAVE__STAT
     /* MS C library seems to define stat and _stat. The definition
      *         is identical. Still, mapping them to each other causes a warning. */
#    ifndef _MSC_VER
#      define stat(x,y) _stat(x,y)
#    endif
#    define HAVE_STAT
#  endif
#endif

xmlParserInputPtr xmlNoNetExternalEntityLoader(const char *URL,
	                                       const char *ID,
					       xmlParserCtxtPtr ctxt);

#define LIBXML_DEBUG_ENABLED
//#define DEBUG

static int debug = 0;
static int novalid = 0;
static int noout = 0;
static int latex = 0;
#ifdef LIBXML_DOCB_ENABLED
static int docbook = 0;
#endif
#ifdef LIBXML_HTML_ENABLED
static int html = 0;
#endif

static const char *params[16 + 1];
static int nbparams = 0;
static const char *output = NULL;

void LatexEscape(xmlNodePtr node);

/*
 *  The LaTeX translation table. Translate unicodes to LaTeX escape sequences.
 */

//#define DEBUG

struct
{
   int           unicode;
   unsigned char *sequence;
} textab[] =
{
   {  '#',     "\\#"              },
   {  '$',     "\\$"              },
   {  '%',     "\\%"              },
   {  '&',     "\\&"              },
   {  '<',     "$<$"              },
   {  '>',     "$>$"              },
   {  '\\',    "$\\backslash$"    },
   {  '^',     "$\\wedge$"        },
   {  '_',     "\\_"              },
   {  '{',     "\\{"              },
   {  '|',     "$\\mid$"          },
   {  '}',     "\\}"              },
   {  '~',     "$\\sim$"          },
   {  160,     "~"                },
   {  161,     "!`"               },
   {  162,     "..."              },
   {  163,     "\\pounds"         },
   {  164,     "..."              },
   {  165,     "..."              },
   {  166,     "..."              },
   {  167,     "\\S"              },
   {  168,     "\\\"{ }"          },
   {  169,     "\\copyright"      },
   {  170,     "..."              },
   {  171,     "..."              },
   {  172,     "$\\neg$"          },
   {  173,     "..."              },
   {  174,     "..."              },
   {  175,     "\\={ }"           },
   {  176,     "$^{\\circ}$"      },
   {  177,     "$\\pm$"           },
   {  178,     "$^{2}$"           },
   {  179,     "$^{3}$"           },
   {  180,     "\\'{ }"           },
   {  181,     "$\\mu$"           },
   {  182,     "\\P"              },
   {  183,     "$\\cdot$"         },
   {  184,     "\\c{ }"           },
   {  185,     "$^{1}$"           },
   {  186,     "..."              },
   {  187,     "..."              },
   {  188,     "$\\frac{1}{4}$"   },
   {  189,     "$\\frac{1}{2}$"   },
   {  190,     "$\\frac{3}{4}$"   },
   {  191,     "?`"               },
   {  192,     "\\`{A}"           },
   {  193,     "\\'{A}"           },
   {  194,     "\\^{A}"           },
   {  195,     "\\~{A}"           },
   {  196,     "\\\"{A}"          },
   {  197,     "\\AA"             },
   {  198,     "\\AE"             },
   {  199,     "\\c{C}"           },
   {  200,     "\\`{E}"           },
   {  201,     "\\'{E}"           },
   {  202,     "\\^{E}"           },
   {  203,     "\\\"{E}"          },
   {  204,     "\\`{I}"           },
   {  205,     "\\'{I}"           },
   {  206,     "\\^{I}"           },
   {  207,     "\\\"{I}"          },
   {  208,     "..."              },
   {  209,     "\\~{N}"           },
   {  210,     "\\`{O}"           },
   {  211,     "\\'{O}"           },
   {  212,     "\\^{O}"           },
   {  213,     "\\~{O}"           },
   {  214,     "\\\"{O}"          },
   {  215,     "$\\times$"        },
   {  216,     "\\O"              },
   {  217,     "\\`{U}"           },
   {  218,     "\\'{U}"           },
   {  219,     "\\^{U}"           },
   {  220,     "\\\"{U}"          },
   {  221,     "\\'{Y}"           },
   {  222,     "..."              },
   {  223,     "\\ss"             },
   {  224,     "\\`{a}"           },
   {  225,     "\\'{a}"           },
   {  226,     "\\^{a}"           },
   {  227,     "\\~{a}"           },
   {  228,     "\\\"{a}"          },
   {  229,     "\\aa"             },
   {  230,     "\\ae"             },
   {  231,     "\\c{c}"           },
   {  232,     "\\`{e}"           },
   {  233,     "\\'{e}"           },
   {  234,     "\\^{e}"           },
   {  235,     "\\\"{e}"          },
   {  236,     "\\`{i}"           },
   {  237,     "\\'{i}"           },
   {  238,     "\\^{i}"           },
   {  239,     "\\\"{i}"          },
   {  240,     "..."              },
   {  241,     "\\~{n}"           },
   {  242,     "\\`{o}"           },
   {  243,     "\\'{o}"           },
   {  244,     "\\^{o}"           },
   {  245,     "\\~{o}"           },
   {  246,     "\\\"{o}"          },
   {  247,     "$\\div$"          },
   {  248,     "\\o"              },
   {  249,     "\\`{u}"           },
   {  250,     "\\'{u}"           },
   {  251,     "\\^{u}"           },
   {  252,     "\\\"{u}"          },
   {  253,     "\\'{y}"           },
   {  254,     "..."              },
   {  255,     "\\\"{y}"          },
   {  338,     "\\OE"             },
   {  339,     "\\oe"             },
   {  352,     "\\u{S}"           },
   {  353,     "\\u{s}"           },
   {  376,     "\\\"{Y}"          },
   {  913,     "A"                },
   {  914,     "B"                },
   {  915,     "$\\Gamma$"        },
   {  916,     "$\\Delta$"        },
   {  917,     "E"                },
   {  918,     "Z"                },
   {  919,     "H"                },
   {  920,     "$\\Theta$"        },
   {  921,     "I"                },
   {  922,     "K"                },
   {  923,     "$\\Lambda$"       },
   {  924,     "M"                },
   {  925,     "N"                },
   {  926,     "$\\Xi$"           },
   {  927,     "O"                },
   {  928,     "$\\Pi$"           },
   {  929,     "P"                },
   {  931,     "$\\Sigma$"        },
   {  932,     "T"                },
   {  933,     "$\\Upsilon$"      },
   {  934,     "$\\Phi$"          },
   {  935,     "X"                },
   {  936,     "$\\Psi$"          },
   {  937,     "$\\Omega$"        },
   {  945,     "$\\alpha$"        },
   {  946,     "$\\beta$"         },
   {  947,     "$\\gamma$"        },
   {  948,     "$\\delta$"        },
   {  949,     "$\\epsilon$"      },
   {  950,     "$\\zeta$"         },
   {  951,     "$\\eta$"          },
   {  952,     "$\\theta$"        },
   {  953,     "$\\iota$"         },
   {  954,     "$\\kappa$"        },
   {  955,     "$\\lambda$"       },
   {  956,     "$\\mu$"           },
   {  957,     "$\\nu$"           },
   {  958,     "$\\xi$"           },
   {  959,     "o"                },
   {  960,     "$\\pi$"           },
   {  961,     "$\\rho$"          },
   {  962,     "$\\varsigma$"     },
   {  963,     "$\\sigma$"        },
   {  964,     "$\\tau$"          },
   {  965,     "$\\upsilon$"      },
   {  966,     "$\\phi$"          },
   {  967,     "$\\chi$"          },
   {  968,     "$\\psi$"          },
   {  969,     "$\\omega$"        },
   { 8224,     "\\dag"            },
   { 8225,     "\\ddag"           },
   { 8226,     "$\\bullet$"       },
   { 8230,     "$\\ldots$"        },
   { 8242,     "$\\prime$"        },
   { 8243,     "$\\prime\\prime$" },
   { 8254,     "\\={ }"           },
   { 8364,     "\\EUR"            },
   { 8465,     "$\\Im$"           },
   { 8472,     "$\\wp$"           },
   { 8476,     "$\\Re$"           },
   { 8482,     "$^{TM}$"          },
   { 8501,     "$\\aleph$"        },
   { 8592,     "$\\leftarrow$"    },
   { 8593,     "$\\uparrow$"      },
   { 8594,     "$\\rightarrow$"   },
   { 8595,     "$\\downarrow$"    },
   { 8596,     "$\\leftrightarrow$" },
   { 8629,     "$\\hookleftarrow$"},
   { 8656,     "$\\Leftarrow$"    },
   { 8657,     "$\\Uparrow$"      },
   { 8658,     "$\\Rightarrow$"   },
   { 8659,     "$\\Downarrow$"    },
   { 8660,     "$\\Leftrightarrow$" },
   { 8704,     "$\\forall$"       },
   { 8706,     "$\\partial$"      },
   { 8707,     "$\\exists$"       },
   { 8709,     "$\\emptyset$"     },
   { 8711,     "$\\nabla$"        },
   { 8712,     "$\\in$"           },
   { 8713,     "..."              },
   { 8715,     "$\\ni$"           },
   { 8719,     "$\\prod$"         },
   { 8721,     "$\\sum$"          },
   { 8722,     "$-$"              },
   { 8727,     "$\\ast$"          },
   { 8730,     "$\\surd$"         },
   { 8733,     "$\\propto$"       },
   { 8734,     "$\\infty$"        },
   { 8736,     "$\\angle$"        },
   { 8743,     "$\\wedge$"        },
   { 8744,     "$\\vee$"          },
   { 8745,     "$\\cap$"          },
   { 8746,     "$\\cup$"          },
   { 8747,     "$\\int$"          },
   { 8756,     "$\\leadsto$"      },
   { 8764,     "$\\sim$"          },
   { 8773,     "$\\cong$"         },
   { 8776,     "$\\approx$"       },
   { 8800,     "$\\neq$"          },
   { 8801,     "$\\equiv$"        },
   { 8804,     "$\\leq$"          },
   { 8805,     "$\\geq$"          },
   { 8834,     "$\\subset$"       },
   { 8835,     "$\\supset$"       },
   { 8836,     "..."              },
   { 8838,     "$\\subseteq$"     },
   { 8839,     "$\\supseteq$"     },
   { 8853,     "$\\oplus$"        },
   { 8855,     "$\\otimes$"       },
   { 8869,     "$\\perp$"         },
   { 8901,     "$\\cdot$"         },
   { 8968,     "$\\lceil$"        },
   { 8969,     "$\\rceil$"        },
   { 8970,     "$\\lfloor$"       },
   { 8971,     "$\\rfloor$"       },
   { 9001,     "$\\langle$"       },
   { 9002,     "$\\rangle$"       },
   { 9674,     "$\\Diamond$"      },
   { 9824,     "$\\spadesuit$"    },
   { 9827,     "$\\clubsuit$"     },
   { 9829,     "$\\heartsuit$"    },
   { 9830,     "$\\diamondsuit$"  },
};

unsigned char *latex_translate(int code)
{
   char   *translation;
   int    i;

   translation = 0;

   for (i = 0; translation == 0 && textab[i].unicode <= code
               && i < sizeof(textab) / sizeof(textab[0]); i++)
   {
      if (textab[i].unicode == code)
      {
         translation = textab[i].sequence;
#ifdef DEBUG
         fprintf(stderr, "\n*** translation for 0x%04x is %s\n", code, translation);
#endif
      }
   }
   return  translation;
}

static void
xsltProcess(xmlDocPtr doc, xsltStylesheetPtr cur, const char *filename) {
    xmlDocPtr res;

    xmlXIncludeProcessFlags(doc, XSLT_PARSE_OPTIONS);

#ifdef DEBUG
    xmlDebugDumpDocument(stdout, doc);
#endif

    if (latex)
    {
       LatexEscape(xmlDocGetRootElement(doc));
    }

    if (output == NULL) {
	res = xsltApplyStylesheet(cur, doc, params);

	xmlFreeDoc(doc);
	if (res == NULL) {
	    fprintf(stderr, "no result for %s\n", filename);
	    return;
	}
	if (noout) {
	    xmlFreeDoc(res);
	    return;
	}
#ifdef LIBXML_DEBUG_ENABLED
	if (debug)
	    xmlDebugDumpDocument(stdout, res);
	else {
#endif
	    if (cur->methodURI == NULL) {
		xsltSaveResultToFile(stdout, res, cur);
	    } else {
		if (xmlStrEqual
		    (cur->method, (const xmlChar *) "xhtml")) {
		    fprintf(stderr, "non standard output xhtml\n");
		    xsltSaveResultToFile(stdout, res, cur);
		} else {
		    fprintf(stderr,
			    "Unsupported non standard output %s\n",
			    cur->method);
		    xsltSaveResultToFile(stdout, res, cur);
		}
	    }
#ifdef LIBXML_DEBUG_ENABLED
	}
#endif

	xmlFreeDoc(res);
    } else {
	xsltRunStylesheet(cur, doc, params, output, NULL, NULL);
	xmlFreeDoc(doc);
    }
}


void LatexEscape(xmlNodePtr node)
{
   xmlNodePtr  child;
   xmlAttrPtr  attribute;
   xmlChar     *byte;
   int         unicode;

   unsigned char  *translation, *replacement;
   int            textlen;
   int            i;

#ifdef DEBUG
      fprintf(stderr, "\n-------- Node type %d\n", node->type);
#endif
   if (xmlNodeIsText(node))
   {
#ifdef DEBUG
      fprintf(stderr, "\n-------- Text node at %p (%d bytes):\n%s",
                      node, strlen(node->content), node->content);
#endif

      textlen = strlen(node->content);
      if (textlen < 100)
      {
         textlen = 100;
      }
      replacement = malloc(textlen);
      i           = 0;
      
      for (byte = node->content; *byte != '\0'; byte++)
      {
         /* transform from utf-8 to unicode */

         if ((*byte & 0x80) == 0)
         {
            unicode = *byte;
         }
         else if ((*byte & 0xe0) == 0xc0)
         {
#ifdef DEBUG
            fprintf(stderr, "2-byte unicode: 0x%02x, 0x%02x\n",byte[0], byte[1]);
#endif
            unicode = (*byte & 0x1f) << 6;
            byte++;
            unicode |= *byte & 0x3f;
#ifdef DEBUG
            fprintf(stderr, "Unicode = 0x%04x\n", unicode);
#endif
         }
         else if ((*byte & 0xf0) == 0xe0)
         {
             /*  3-byte unicode */

            unicode = (*byte & 0x0f) << 12;
            byte++;
            unicode |= (*byte & 0x3f) << 6;
            byte++;
            unicode |= *byte & 0x3f;
         }
         else
         {
            fprintf(stderr, "More than 3 bytes utf-8 is not supported.\n");
         }

         translation = latex_translate(unicode);

         /*  Append the original byte or the translation to the replacment text */

         if (translation == 0)
         {
            if (i + 1 >= textlen)
            {
               textlen += 100;
               replacement = realloc(replacement, textlen);
            }
            replacement[i++] = *byte;
         }
         else
         {
            int   j;

            if (i + strlen(translation) + 1 >= textlen)
            {
               textlen += 100;
               replacement = realloc(replacement, textlen);
            }
            for (j = 0; translation[j]; j++)
            {
               replacement[i++] = translation[j];
            } 
         }
      }
      replacement[i] = '\0';
      xmlNodeSetContent(node, replacement);
#ifdef DEBUG
      fprintf(stderr, "\n-------- Replacement text:\n%s", replacement);
#endif
      free(replacement);
   }
   else
   {
      for (attribute = node->properties; attribute != 0; attribute = attribute->next)
      {
         LatexEscape(attribute->children);
      }
      for (child = node->children; child != 0; child = child->next)
      {
         if (strcmp(child->name, "verbatim") != 0 && strcmp(child->name, "example") != 0)
         {
            LatexEscape(child);
         }
      }
   }
}

static void usage(const char *name)
{
    printf("Usage: %s [options] stylesheet file [file ...]\n", name);
    printf("   Options:\n");
    printf("\t--version or -V: show the version of libxml and libxslt used\n");
    printf("\t--verbose or -v: show logs of what's happening\n");
    printf("\t--output file or -o file: save to a given file\n");
    printf("\t--debug: dump the tree of the result instead\n");
    printf("\t--latex: transform special characters for latex\n");
    printf("\t--novalid: skip the Dtd loading phase\n");
    printf("\t--noout: do not dump the result\n");
    printf("\t--maxdepth val : increase the maximum depth\n");
#ifdef LIBXML_HTML_ENABLED
    printf("\t--html: the input document is(are) an HTML file(s)\n");
#endif
#ifdef LIBXML_DOCB_ENABLED
    printf("\t--docbook: the input document is SGML docbook\n");
#endif
    printf("\t--param name value : pass a (parameter,value) pair\n");
    printf("\t      string values must be quoted like \"'string'\"\n");
    printf("\t--nonet refuse to fetch DTDs or entities over network\n");
#ifdef LIBXML_CATALOG_ENABLED
    printf("\t--catalogs : use SGML catalogs from $SGML_CATALOG_FILES\n");
    printf("\t             otherwise XML Catalogs starting from \n");
    printf("\t         file:///etc/xml/catalog are activated by default\n");
#endif
}

int
main(int argc, char **argv)
{
    int i;
    xsltStylesheetPtr cur = NULL;
    xmlDocPtr doc, style;

    if (argc <= 1) {
        usage(argv[0]);
        return (1);
    }

    xmlInitMemory();

    LIBXML_TEST_VERSION

    xmlLineNumbersDefault(1);

    if (novalid == 0)
        xmlLoadExtDtdDefaultValue = XML_DETECT_IDS | XML_COMPLETE_ATTRS;
    else
        xmlLoadExtDtdDefaultValue = 0;
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "-"))
            break;

        if (argv[i][0] != '-')
            continue;
#ifdef LIBXML_DEBUG_ENABLED
        if ((!strcmp(argv[i], "-debug")) || (!strcmp(argv[i], "--debug"))) {
            debug++;
        } else
#endif
        if ((!strcmp(argv[i], "-v")) ||
                (!strcmp(argv[i], "-verbose")) ||
                (!strcmp(argv[i], "--verbose"))) {
            xsltSetGenericDebugFunc(stderr, NULL);
        } else if ((!strcmp(argv[i], "-o")) ||
                   (!strcmp(argv[i], "-output")) ||
                   (!strcmp(argv[i], "--output"))) {
            i++;
            output = argv[i++];
        } else if ((!strcmp(argv[i], "-V")) ||
                   (!strcmp(argv[i], "-version")) ||
                   (!strcmp(argv[i], "--version")))
       	{

        } else if ((!strcmp(argv[i], "-novalid")) ||
                   (!strcmp(argv[i], "--novalid"))) {
            novalid++;
        } else if ((!strcmp(argv[i], "-latex")) ||
                   (!strcmp(argv[i], "--latex"))) {
            latex++;
        } else if ((!strcmp(argv[i], "-noout")) ||
                   (!strcmp(argv[i], "--noout"))) {
            noout++;
#ifdef LIBXML_HTML_ENABLED
        } else if ((!strcmp(argv[i], "-html")) ||
                   (!strcmp(argv[i], "--html"))) {
            html++;
#endif
        } else if ((!strcmp(argv[i], "-nonet")) ||
                   (!strcmp(argv[i], "--nonet"))) {
            xmlSetExternalEntityLoader(xmlNoNetExternalEntityLoader);
#ifdef LIBXML_CATALOG_ENABLED
        } else if ((!strcmp(argv[i], "-catalogs")) ||
                   (!strcmp(argv[i], "--catalogs"))) {
            const char *catalogs;

            catalogs = getenv("SGML_CATALOG_FILES");
            if (catalogs == NULL) {
                fprintf(stderr, "Variable $SGML_CATALOG_FILES not set\n");
            } else {
                xmlLoadCatalogs(catalogs);
            }
#endif
        } else if ((!strcmp(argv[i], "-param")) ||
                   (!strcmp(argv[i], "--param"))) {
            i++;
            params[nbparams++] = argv[i++];
            params[nbparams++] = argv[i];
            if (nbparams >= 16) {
                fprintf(stderr, "too many params\n");
                return (1);
            }
        } else if ((!strcmp(argv[i], "-maxdepth")) ||
                   (!strcmp(argv[i], "--maxdepth"))) {
            int value;

            i++;
            if (sscanf(argv[i], "%d", &value) == 1) {
                if (value > 0)
                    xsltMaxDepth = value;
            }
        } else {
            fprintf(stderr, "Unknown option %s\n", argv[i]);
            usage(argv[0]);
            return (1);
        }
    }
    params[nbparams] = NULL;

    xsltSetXIncludeDefault(1);
    /*
     * Replace entities with their content.
     */
    xmlSubstituteEntitiesDefault(1);

    /*
     * Register the EXSLT extensions and the test module
     */
    exsltRegisterAll();
    //xsltRegisterTestModule();

    for (i = 1; i < argc; i++) {
        if ((!strcmp(argv[i], "-maxdepth")) ||
            (!strcmp(argv[i], "--maxdepth"))) {
            i++;
            continue;
        } else if ((!strcmp(argv[i], "-o")) ||
                   (!strcmp(argv[i], "-output")) ||
                   (!strcmp(argv[i], "--output"))) {
            i++;
	    continue;
	}
        if ((!strcmp(argv[i], "-param")) || (!strcmp(argv[i], "--param"))) {
            i += 2;
            continue;
        }
        if ((argv[i][0] != '-') || (strcmp(argv[i], "-") == 0)) {
	    style = xmlParseFile((const char *) argv[i]);
	    if (style == NULL) {
		fprintf(stderr,  "cannot parse %s\n", argv[i]);
		cur = NULL;
	    } else {
		cur = xsltLoadStylesheetPI(style);
		if (cur != NULL) {
		    /* it is an embedded stylesheet */
		    xsltProcess(style, cur, argv[i]);
		    xsltFreeStylesheet(cur);
		    goto done;
		}
		cur = xsltParseStylesheetDoc(style);
		if (cur != NULL) {
		    if (cur->indent == 1)
			xmlIndentTreeOutput = 1;
		    else
			xmlIndentTreeOutput = 0;
		    i++;
		} else {
		    xmlFreeDoc(style);
		    goto done;
		}
	    }
            break;

        }
    }

    /*
     * disable CDATA from being built in the document tree
     */
    xmlDefaultSAXHandlerInit();
    xmlDefaultSAXHandler.cdataBlock = NULL;

    if ((cur != NULL) && (cur->errors == 0)) {
        for (; i < argc; i++) {
	    doc = NULL;
#ifdef LIBXML_HTML_ENABLED
            if (html)
                doc = htmlParseFile(argv[i], NULL);
            else
#endif
                doc = xmlParseFile(argv[i]);
            if (doc == NULL) {
                fprintf(stderr, "unable to parse %s\n", argv[i]);
                continue;
            }
	    xsltProcess(doc, cur, argv[i]);
        }
    }
    if (cur != NULL)
        xsltFreeStylesheet(cur);
done:
    //xsltCleanupGlobals();
    xmlCleanupParser();
    xmlMemoryDump();
    return (0);
}

